#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define IMIN(a, b) ((a) < (b) ? (a) : (b))

double wtime()
{
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec * 1E-9;
}

void sum(double *a, double *b, int m, int n)
{
    for (int i = 0; i < m; i++) {
        for (int j = 0; j < n; j++)
            a[i * n + j] += b[j * n + i];
    }
}

void sum_block(double *a, double *b, int m, int n)
{
    int bs = 8;
    for (int i = 0; i < m; i += bs) {
        for (int j = 0; j < n; j += bs) {
            for (int ii = i; ii < IMIN(i + bs, m); ii++) {
                for (int jj = j; jj < IMIN(j + bs, n); jj++) {
                    a[ii * n + jj] += b[jj * n + ii];
                }
            }
        }
    }
}

double run(int m, int n)
{
    double *a, *b;
    a = aligned_alloc(64, sizeof(*a) * m * n);
    b = aligned_alloc(64, sizeof(*b) * m * n);
    for (int i = 0; i < m; i++) {
        for (int j = 0; j < n; j++) {
            a[i * n + j] = 0;
            b[i * n + j] = i + j;
        }
    }

    double t = wtime();
    sum(a, b, m, n);
    //sum_block(a, b, m, n);
    t = wtime() - t;

    free(a);
    free(b);
    return t;
}

int main()
{
    int nreps = 100;
    int n = 1000;
    printf("# [m]  [time, sec]   [MFLOPS]\n");
    for (int m = 1000; m <= 10000; m += 1000) {
        double t = 0;
        for (int i = 0; i < nreps; i++)
             t += run(m, n);
        t /= nreps;
        printf(" %-6d  %-12.6f  %-12.2f\n", m, t, 1.0 * m * n * 1e-6 / t);
    }

    return 0;
}

